	.286c

	PAGE	80,128

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;	File:	nice.asm
;	Vers:	1.0
;	Date:	23-oct-1998
;	Auth:	PJD
;
;	Description:
;	~~~~~~~~~~~~
;	Tiny TSR which releases time back to the CPU when running a DOS
;	app under Windows (3.1 or W95) and which app polls 'stdin' (ie.
;	the keyboard) using DOS function 6 (raw I/O).
;
;	It does this by trapping all DOS calls through int 21h, and on
;	any call where AH = 06h and DL = FFh makes a call to the multiplex
;	interrupt with AX = 1680h (see Ralf Brown's List).
;
;	Amendments:
;	1.0	23-oct-98	PJD	Initial version of file.
;	1.1	27-oct-98	PJD	Added /U uninstall option and tidied
;					up code a little (not that anyone
;					would notice...)
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
VMAJOR		equ	'1'                     ;major/minor version numbers
VMINOR		equ	'1'                     ;
						;
DOS		equ	21h			;DOS interrupt number
MULTIPLEX	equ	2Fh			;Multiplex interrupt
						;
SIG_AX		equ	0CEDEh			;unused, and aposite...
SIG_1		equ	0BADDh			;
SIG_2		equ	0C0DEh			;...as is this :-)
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;	MACROS
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
jmpf	MACRO					; JMPF instruction
	db	0EAh				; "jump far immediate" opcode
	ENDM					;
						;
callf	MACRO					; CALLF instruction
	db	09Ah				; "call far immediate" opcode
	ENDM					;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
CODE	SEGMENT PUBLIC	'CODE'                  ;
	ASSUME	CS:CODE, DS:NOTHING, ES:NOTHING ;
						;
	ORG	2Ch				;loc of env seg pointer
						;
envseg	LABEL	WORD				;
						;
	ORG	80h				;
						;
cmdtail LABEL	BYTE				;
						;
	ORG	100h				;
						;
inflag	db	?				;re-entrancy flag
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
start	PROC	NEAR				;
						;
	ORG	100h				;rewind instruction counter
						;
	jmp	short	initialise		;
						;
start	ENDP					;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
	EVEN					;
						;
DOS_ISR PROC	NEAR				;
						;
	test	cs:[inflag], 0ffh		;check re-entrancy...
	jnz	old_int 			;
						;
	cmp	ah, 06h 			;DOS raw I/O service?
	jnz	old_int 			;no - ignore
						;
	cmp	dl, 0FFh			;polled input?
	jnz	old_int 			;no - ignore
						;
	inc	cs:[inflag]			;bolt the door
	pusha					;release time slice
	mov	ax, 1680h			;
	int	MULTIPLEX			;
	popa					;
	dec	cs:[inflag]			;unbolt the door
						;
old_int:					;
	jmpf					;chain original vector
oldv_off	dw	0			;original offset addr of INT 21h vector
oldv_seg	dw	0			;original segment addr of INT 21h vector
						;
DOS_ISR ENDP					;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
	EVEN					;
						;
	dw	SIG_1				;2 "signature" words
	dw	SIG_2				;(immediately below our int 2f)
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
int2f	PROC	FAR				;
						;
	pushf					;
	cmp	ax, SIG_AX			;is it ours?
	jnz	not_ours			;no
						;
	cmp	si, offset int2f		;ES:SI should be pointing at
	jnz	not_ours			;  the transient int2f
						;
	cmp	word ptr es:-4[si], SIG_1	;so check the signature bytes
	jnz	not_ours			;
						;
	cmp	word ptr es:-2[si], SIG_2	;
	jnz	not_ours			;
						;
	push	cs				;
	pop	es				;point ES at the resident copy
	not	ax				;and add another clue
	popf					;
	iret					;
						;
not_ours:					;
	popf					;
	jmpf					;chain original vector
old_2f_off	dw	5678h			;these filled in by 'initialise'
old_2f_seg	dw	1234h			;
						;
int2f	ENDP					;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
initialise	PROC	NEAR			;
						; set up local stack
	mov	ax, cs				;
	mov	ss, ax				;
	mov	sp, offset init_stack		;
	mov	ds, ax				; set DS = CS
						;
	ASSUME	DS:CODE 			;
						;
	call	signon				;
						;
	mov	ax, 1680h			;see if call is supported
	int	MULTIPLEX			;
	or	al, al				;zero if OK
	jz	proceed 			;
						;exit if not
	push	1				;set exit code
	call	display_and_exit		;
	db	'Unsupported DPMI call - not installed.$'
						;
proceed:					;
	call	check_cmdtail			;
	mov	ax, SIG_AX			;installation check
	mov	si, offset int2f		;point ES:SI at transient int2f
	push	cs				;
	pop	es				;
	int	MULTIPLEX			;installation check
	cmp	ax, (NOT SIG_AX)		;
	jnz	OK_to_install			;no - ok to install
						;else might already be resident
	mov	ax, es				;did ES change?
	mov	bx, cs				;
	cmp	ax, bx				;
	jz	OK_to_install			;no - ok to install
						;
	cmp	es:-4[si], SIG_1		;double-check signature bytes
	jnz	OK_to_install			;different
						;
	cmp	es:-2[si], SIG_2		;
	jnz	OK_to_install			;different
						;
	mov	ds:[installed_seg], es		;pretty conclusive!
						;
	cmp	byte ptr ds:[uninstall_flg], 1	;are we uninstalling?
	jnz	already_resident		;no - therefore an error
						;
	call	remove				;try to uninstall
	jz	removed_OK			;successful
						;
	push	3				;set exit code
	call	display_and_exit		;vectors have been changed...
	db	'Cannot uninstall.$'            ;
						;
removed_OK:					;
	push	0				;all OK
	call	display_and_exit		;
	db	'Successfully uninstalled.$'    ;
						;
already_resident:				;
	push	2				;set exit code
	call	display_and_exit		;
	db	'Already resident.$'            ;
						;
OK_to_install:					;NICE is not resident
	cmp	byte ptr ds:[uninstall_flg], 1	;so we can't uninstall it...
	jnz	continue			;
						;attempt to uninstall
	push	4				; when not resident...
	call	display_and_exit		;
	db	'Nothing to uninstall!$'        ;
						;
continue:					;
	mov	ax, 352Fh			;read current int 2F vector
	int	DOS				;
	mov	ds:[old_2f_off], bx		;save current vector to chain
	mov	ds:[old_2f_seg], es		;
	mov	ax, 252Fh			;hook multiplex interrupt
	mov	dx, offset int2f		;
	int	DOS				;
						;
	mov	ds:[inflag], 0FFh		;bolt the door till we're installed
	push	es				;save ES
	mov	ax, 3521h			;get int 21 vec
	int	DOS				;
	mov	ds:[oldv_off], bx		;save it
	mov	ds:[oldv_seg], es		;
						;
	sub	ax, ax				;hook int 21 manually
	mov	es, ax				;set ES = 0
	mov	di, DOS * 4			;set offset into vector table
	pushf					;
	push	cs				;prepare CS for planting
	mov	ax, offset DOS_ISR		;
	cld					;auto incrememt
	cli					;ints off while patching vector
	stosw					;patch offset
	pop	ax				;retrieve CS in AX
	stosw					;patch segment
	mov	ds:[inflag], 0			;open the floodgate...
	popf					;ints back on
	pop	es				;restore ES
						;all done
	mov	es, ds:[envseg] 		;release environment block
	mov	ah, 49h 			;
	int	DOS				;
						;
	call	exit_OK 			;
	db	'Successfully installed.$'      ;
exit_OK:					;
	pop	dx				;
	mov	ah, 9				;
	int	DOS				;
						;
	mov	dx, offset initialise		;release unneeded memory
	shr	dx, 4				;convert to para address
	inc	dx				;just in case
	mov	ax, 3100h			;TSR function, return 0
	int	DOS				;
						;
initialise	ENDP				;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
signon	PROC	NEAR				;
						;
	mov	dx, offset prog 		;show name and version
	mov	ah, 9				;
	int	DOS				;
	ret					;
						;
signon	ENDP					;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
display_and_exit  PROC	NEAR			;
						;
	pop	dx				;address of string to display
	mov	ah, 9				;
	int	DOS				;display it
	pop	ax				;retrieve exit code
	mov	ah, 4Ch 			;and exit
	int	DOS				;
	jmp	short	$			;just in case...
						;
display_and_exit  ENDP				;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
check_cmdtail	PROC	NEAR			;
						;
	mov	si, offset cmdtail		;look for -u or /u
	cld					;just in case
	lodsb					;
	or	al, al				;
	jz	pct_ret 			;no commandtail
	lodsb					;else get first byte
getchar:					;
	cmp	al, 0Dh 			;end of commandtail?
	jz	pct_ret 			;yes
	cmp	al, '-'                         ;
	jz	check_next			;
	cmp	al, '/'                         ;
	jz	check_next			;
	lodsb					;
	jmp	short	getchar 		;
check_next:					;
	lodsb					;
	cmp	al, 'U'                         ;
	jz	U_found 			;
	cmp	al, 'u'                         ;
	jnz	getchar 			;
U_found:					;
	inc	byte ptr ds:[uninstall_flg]	;
pct_ret:					;
	ret					;
						;
check_cmdtail	ENDP				;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
						;
remove	PROC	NEAR				;
						;
	mov	ax, 352Fh			;read current int 2F vector
	int	DOS				;
	mov	ax, es				;
	cmp	ax, ds:[installed_seg]		;has it changed?
	jnz	cant_remove			;yes - zflag clear
						;
	mov	ax, 3521h			;read current int 21 vector
	int	DOS				;
	mov	ax, es				;
	cmp	ax, ds:[installed_seg]		;has it changed?
	jnz	cant_remove			;yes - zflag clear
						;
	push	ds				;OK to restore original vectors
	push	es				;save just in case
	lds	dx, dword ptr es:[old_2f_off]	;
	mov	ax, 252Fh			;reset int 2F
	int	DOS				;
						;
	pop	es				;reset to resident seg
	lds	dx, dword ptr es:[oldv_off]	;
	mov	ax, 2521h			;reset int 21
	int	DOS				;
	pop	ds				;restore DS
						;
	mov	es, ds:[installed_seg]		;the resident program segment
	mov	ah, 49h 			;free its allocated memory
	int	21h				;
						;
	xor	ax, ax				;set zflag
						;
cant_remove:					;
	ret					;
						;
remove	ENDP					;
						;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	EVEN

_is		dw	100 dup (0)

init_stack	LABEL	NEAR		;local stack here

installed_seg	dw	0
uninstall_flg	db	0		;assume we want to install

prog		db	'NICE v', VMAJOR, '.', VMINOR, 0Dh, 0Ah, '$'

CODE	ENDS

	END

